Attribute VB_Name = "modPickFolderDialog"
' Autor: F. Schler (frank@activevb.de)
' Datum: 01/2021

Option Explicit

' ----==== Const ====----
Private Const S_OK As Long = &H0&
Private Const FOS_PICKFOLDERS = &H20
Private Const SIGDN_FILESYSPATH = &H80058000
Private Const CC_STDCALL As Long = &H4&
Private Const IID_Release As Long = &H8&
Private Const CLSCTX_INPROC As Long = &H1&
Private Const CLSID_FileOpenDialog As String = "{DC1C5A9C-E88A-4DDE-A5A1-60F82A20AEF7}"
Private Const IID_IFileDialog As String = "{42F85136-DB7E-439C-85F1-E4075D135FC8}"
Private Const IID_IShellItem As String = "{43826D1E-E718-42EE-BC55-A1E261C37BFE}"

' ----==== Enums ====----
Private Enum vtb_Interfaces
    
    ' IModalWindow
    Show_ = 3
    
    ' IFileDialog
    SetOptions_ = 9
    GetOptions_ = 10
    SetFolder_ = 12
    SetTitle_ = 17
    GetResult_ = 20

    'IShellItem
    GetDisplayName_ = 5

End Enum

' ----==== Type ====----
Private Type GUID
    data1 As Long
    data2 As Integer
    data3 As Integer
    data4(0 To 7) As Byte
End Type

' ----==== Kernel32.dll Deklarationen ====----
Private Declare Sub CopyMemory Lib "Kernel32.dll" _
                    Alias "RtlMoveMemory" ( _
                    ByRef pDst As Any, _
                    ByRef pSrc As Any, _
                    ByVal ByteLen As Long)
                    
Private Declare Function lstrlenW Lib "Kernel32.dll" ( _
                         ByVal lpString As Long) As Long
                         
' ----==== Shell32.dll Deklarationen ====----
Private Declare Function SHCreateItemFromParsingName Lib "Shell32.dll" ( _
                         ByVal pszPath As Long, _
                         ByVal pbc As Long, _
                         ByVal riid As Long, _
                         ByRef ppv As Long) As Long
                         
' ----==== Ole32.dll Deklarationen ====----
Private Declare Sub CoTaskMemFree Lib "Ole32.dll" ( _
                    ByVal hMem As Long)
                    
Private Declare Function CLSIDFromString Lib "Ole32.dll" ( _
                         ByVal pString As Long, _
                         ByRef pCLSID As GUID) As Long
                         
Private Declare Function CoCreateInstance Lib "Ole32.dll" ( _
                         ByVal rclsid As Long, _
                         ByVal pUnkOuter As Long, _
                         ByVal dwClsContext As Long, _
                         ByVal riid As Long, _
                         ByRef ppv As Long) As Long
                         
' ----==== Oleaut32.dll Deklarationen ====----
Private Declare Function DispCallFunc Lib "Oleaut32.dll" ( _
                         ByVal pvInstance As Long, _
                         ByVal oVft As Long, _
                         ByVal cc As Long, _
                         ByVal vtReturn As VbVarType, _
                         ByVal cActuals As Long, _
                         ByVal prgvt As Any, _
                         ByVal prgpvarg As Any, _
                         ByRef pvargResult As Variant) As Long
                         
' ----==== Variablen ====----
Private m_pIPickFolderDialog As Long
                         
Public Function CreatePickFolderDialog() As Boolean

    Dim bolRet As Boolean
    Dim lngOptions As Long
    
    bolRet = False
    
    If CoCreateInstance(VarPtr(Str2Guid(CLSID_FileOpenDialog)), 0&, _
        CLSCTX_INPROC, VarPtr(Str2Guid(IID_IFileDialog)), _
        m_pIPickFolderDialog) = S_OK Then
        
        If Invoke(m_pIPickFolderDialog, GetOptions_, VarPtr(lngOptions)) = S_OK Then
    
            lngOptions = lngOptions Or FOS_PICKFOLDERS
            
            If Invoke(m_pIPickFolderDialog, SetOptions_, lngOptions) = S_OK Then
    
                bolRet = True
        
            End If
        
        End If
        
    End If
    
    CreatePickFolderDialog = bolRet
    
End Function

Public Function DestroyPickFolderDialog() As Boolean

    DestroyPickFolderDialog = Release(m_pIPickFolderDialog)
    
End Function

Public Function SetPickFolderDialogTitle(Optional ByVal sTitle As _
    String = vbNullChar) As Boolean

    Dim bolRet As Boolean
    
    bolRet = False
    
    If Invoke(m_pIPickFolderDialog, SetTitle_, StrPtr(sTitle)) = S_OK Then

        bolRet = True

    End If
    
    SetPickFolderDialogTitle = bolRet

End Function

Public Function SetPickFolderDialogStartFolder(Optional ByVal sStartFolder As _
    String = vbNullChar) As Boolean
    
    Dim bolRet As Boolean
    Dim pIShellItem As Long
    
    bolRet = False
    
    If SHCreateItemFromParsingName(StrPtr(sStartFolder), 0&, _
        VarPtr(Str2Guid(IID_IShellItem)), pIShellItem) = S_OK Then
        
        If Invoke(m_pIPickFolderDialog, SetFolder_, pIShellItem) = S_OK Then
        
            bolRet = True
            
        End If
        
        Call Release(pIShellItem)
        
    End If
    
    SetPickFolderDialogStartFolder = bolRet
    
End Function

Public Function ShowPickFolderDialog(Optional ByVal hwndOwner As Long = 0&) As Boolean

    Dim bolRet As Boolean
    
    bolRet = False
    
    If Invoke(m_pIPickFolderDialog, Show_, hwndOwner) = S_OK Then
    
        bolRet = True
        
    End If
        
    ShowPickFolderDialog = bolRet
    
End Function

Public Function GetPickFolderDialogResult() As String

    Dim strRet As String
    Dim pszName As Long
    Dim pIShellItem As Long
    
    If Invoke(m_pIPickFolderDialog, GetResult_, VarPtr(pIShellItem)) = S_OK Then
    
        If Invoke(pIShellItem, GetDisplayName_, SIGDN_FILESYSPATH, _
            VarPtr(pszName)) = S_OK Then
            
            strRet = StrPtr2Str(pszName)
            
        End If
        
        Call Release(pIShellItem)
        
    End If
    
    GetPickFolderDialogResult = strRet
    
End Function

Private Function Invoke(ByVal pInterface As Long, _
    ByVal vtb As vtb_Interfaces, ParamArray var()) As Variant
    
    If pInterface <> 0& Then
    
        Invoke = OleInvoke(pInterface, vtb, var)
        
    End If
    
End Function

Private Function Release(ByRef pInterface As Long) As Boolean

    Dim bolRet As Boolean
    Dim vRet As Variant

    bolRet = False
    
    If pInterface <> 0& Then

        If DispCallFunc(pInterface, IID_Release, CC_STDCALL, vbLong, _
            0&, 0&, 0&, vRet) = S_OK Then

            pInterface = 0&
            
            bolRet = True

        End If

    End If
    
    Release = bolRet
    
End Function

Private Function OleInvoke(ByVal pInterface As Long, _
    ByVal lCmd As Long, ParamArray aParam()) As Variant
    
    Dim lItem As Long
    Dim lCount As Long
    Dim vRet As Variant
    Dim vParam As Variant
    Dim olePtr(10) As Long
    Dim oleTyp(10) As Integer

    If pInterface <> 0& Then

        If UBound(aParam) >= 0 Then

            vParam = aParam

            If IsArray(vParam) Then vParam = vParam(0)

            lCount = UBound(vParam)

            For lItem = 0 To lCount

                oleTyp(lItem) = VarType(vParam(lItem))

                olePtr(lItem) = VarPtr(vParam(lItem))

            Next

        End If

        If DispCallFunc(pInterface, lCmd * 4, CC_STDCALL, vbLong, _
            lItem, VarPtr(oleTyp(0)), VarPtr(olePtr(0)), vRet) <> S_OK Then
            
            Debug.Print "Fehler beim Aufrufen der Interface-Funktion!"

        End If

    End If

    OleInvoke = vRet

End Function

Private Function Str2Guid(ByVal str As String) As GUID

    Call CLSIDFromString(StrPtr(str), Str2Guid)
    
End Function

Private Function StrPtr2Str(ByVal lpStrPointer As Long) As String

    Dim lLen As Long
    Dim bBuffer() As Byte

    If lpStrPointer = 0 Then Exit Function

    lLen = lstrlenW(lpStrPointer) * 2& - 1&

    If lLen > 0& Then

        ReDim bBuffer(lLen)

        Call CopyMemory(bBuffer(0), ByVal lpStrPointer, lLen)
        
        Call CoTaskMemFree(lpStrPointer)

        StrPtr2Str = bBuffer

    End If

End Function

